from pathlib import Path
from typing import Optional

import networkx as nx
import pandas as pd

from plasnet.base_graph import BaseGraph
from plasnet.communities import Communities
from plasnet.community_graph import CommunityGraph
from plasnet.utils import DistanceTags


class PlasmidGraph(BaseGraph):
    """
    Class to represent a plasmid graph.
    It represents a full plasmid graph, not partitioned into communities or subcommunities.
    Each node is a plasmid, and each edge represents an abstract distance between two plasmids.
    """

    def __init__(self, graph: Optional[nx.Graph] = None, label: str = "") -> None:
        super().__init__(graph, label)

    @staticmethod
    def build(
        plasmids_filepath: Path, distance_filepath: Path, distance_threshold: float
    ) -> "PlasmidGraph":
        """
        Creates a plasmid graph from plasmid and distance files.

        The plasmid file is a tab-separated file with one column describing all plasmids in the dataset.
        Example of such file:
        plasmid
        AP024796.1
        AP024825.1
        CP012142.1
        CP014494.1
        CP019149.1
        CP021465.1
        CP022675.1
        CP024687.1
        CP026642.1
        CP027485.1

        The distance file is a tab-separated file with 3 columns: plasmid_1, plasmid_2, distance.
        plasmid_1 and plasmid_2 are plasmid names, and distance is a float between 0 and 1.
        The distance threshold is the minimum distance value for two plasmids to be considered connected.

        Example of such file:
        plasmid_1       plasmid_2       distance
        AP024796.1      AP024825.1      0.8
        AP024796.1      CP012142.1      0.5
        AP024796.1      CP014494.1      0.3
        AP024796.1      CP019149.1      0.0
        AP024796.1      CP021465.1      0.0
        AP024796.1      CP022675.1      1.0
        AP024796.1      CP024687.1      0.0
        AP024796.1      CP026642.1      0.5
        AP024796.1      CP027485.1      0.8
        """  # noqa: E501
        plasmids = pd.read_csv(plasmids_filepath)

        distance_df = pd.read_csv(distance_filepath, sep="\t")
        distance_df[DistanceTags.SplitDistanceTag.value] = distance_df["distance"]

        # apply distance threshold
        distance_df = distance_df[
            distance_df[DistanceTags.SplitDistanceTag.value] <= distance_threshold
        ]

        # round distance to 2 decimals
        distance_df[DistanceTags.SplitDistanceTag.value] = distance_df[
            DistanceTags.SplitDistanceTag.value
        ].round(2)

        # create graph
        graph = nx.from_pandas_edgelist(
            distance_df,
            source="plasmid_1",
            target="plasmid_2",
            edge_attr=DistanceTags.SplitDistanceTag.value,
            create_using=PlasmidGraph,
        )
        graph.add_nodes_from(plasmids["plasmid"])

        return PlasmidGraph(graph)

    def split_graph_into_communities(
        self, bh_connectivity: int, bh_neighbours_edge_density: float
    ) -> Communities:
        return Communities(
            CommunityGraph(
                self.subgraph(component),
                bh_connectivity,
                bh_neighbours_edge_density,
                label=f"community_{idx}",
            )
            for idx, component in enumerate(nx.connected_components(self))
        )

    def _get_libs_relative_path(self) -> str:
        return "."

    def _get_samples_selectors_HTML(self) -> str:
        return ""

    def _get_filters_HTML(self) -> str:
        return ""

    def _get_custom_buttons_HTML(self) -> str:
        return ""



from typing import Optional

import networkx as nx
import pandas as pd

from plasnet.ColorPicker import ColorPicker
from plasnet.subcommunity_graph import SubcommunityGraph


class SampleGraph(SubcommunityGraph):
    def __init__(
        self,
        graph: Optional[nx.Graph] = None,
        blackhole_connectivity_threshold: int = 0,
        edge_density: float = 0.0,
        label: str = "",
        colour: str = ColorPicker.get_default_color(),
        sample_plasmid: Optional[pd.DataFrame] = None,
    ):
        super().__init__(graph, blackhole_connectivity_threshold, edge_density, label, colour)
        if sample_plasmid is not None:
            self._sample_to_plasmids = (
                sample_plasmid.groupby("sample")["plasmid"].apply(list).to_dict()
            )

    @classmethod
    def from_subcommunity_graph(
        cls, subcommunity_graph: SubcommunityGraph, sample_plasmid: pd.DataFrame
    ) -> "SampleGraph":
        return cls(
            subcommunity_graph,
            subcommunity_graph._blackhole_connectivity_threshold,
            subcommunity_graph._edge_density,
            subcommunity_graph.label,
            subcommunity_graph._colour,
            sample_plasmid,
        )

    def get_number_of_hit_samples(self) -> int:
        return len(self._sample_to_plasmids)

    def _get_samples_selectors_HTML(self) -> str:
        samples_selectors = []
        for sample, plasmids in self._sample_to_plasmids.items():
            induced_components = self.get_induced_components(plasmids)
            for component_index, component in enumerate(induced_components):
                node_selector = [f"node#{node}" for node in component]
                if component_index == 0:
                    samples_selectors.append(f"sample_selector_nodes['{sample}'] = [];")
                node_selector_str = ", ".join(node_selector)
                samples_selectors.append(
                    f"sample_selector_nodes['{sample}'].push(cy.elements('{node_selector_str}'));"
                )
        samples_selectors_str = "\n".join(samples_selectors)
        return samples_selectors_str

    def _get_filters_HTML(self) -> str:
        previous_filters = super()._get_filters_HTML()

        sample_hits_checkboxes = []
        for sample_index, sample in enumerate(self._sample_to_plasmids):
            colour = ColorPicker.get_color_given_index(sample_index)
            sample_hits_checkboxes.append(
                f'<input type="checkbox" id="{sample}" name="{sample}" '
                f"onclick=\"show_sample_hits('{sample}', '{colour}')\">"
                f'<label for="{sample}">{sample} ({len(self._sample_to_plasmids[sample])} hits) '
                f'<span style="color:{colour}">&#9632;</span></label><br/>'
            )

        html_filters = (
            f"{previous_filters}<br/>"
            f"Show hits for samples:<br/>"
            f"{''.join(sample_hits_checkboxes)}"
        )
        return html_filters

    @property
    def description(self) -> str:
        description = super().description
        description += f" - {self.get_number_of_hit_samples()} samples hit"
        return description



import shutil
from pathlib import Path

from plasnet.base_graph import BaseGraph
from plasnet.blackhole_graph import BlackholeGraph
from plasnet.communities import Communities
from plasnet.list_of_graphs import ListOfGraphs
from plasnet.Templates import Templates
from plasnet.utils import get_libs_dir


class OutputProducer:
    @staticmethod
    def produce_graph_visualisation(graph: BaseGraph, html_path: Path) -> None:
        outdir = html_path.parent
        outdir.mkdir(exist_ok=True, parents=True)

        html = graph.produce_visualisation()
        html_path.write_text(html)

        OutputProducer.copy_libs(outdir)

    @staticmethod
    def produce_communities_visualisation(communities: Communities, outdir: Path) -> None:
        OutputProducer._write_html_for_all_subgraphs(communities, outdir)
        OutputProducer._produce_index_file(outdir, communities, "Communities")

    @staticmethod
    def produce_subcommunities_visualisation(
        subcommunities: ListOfGraphs[BlackholeGraph], outdir: Path
    ) -> None:
        OutputProducer._write_html_for_all_subgraphs(subcommunities, outdir)
        OutputProducer._produce_index_file(outdir, subcommunities, "subcommunity")

    @classmethod
    def _write_html_for_all_subgraphs(
        cls, subgraphs: ListOfGraphs[BaseGraph], outdir: Path
    ) -> None:
        graphs_dir = outdir / "graphs"
        graphs_dir.mkdir(exist_ok=True, parents=True)

        for subgraph_index, subgraph in enumerate(subgraphs):
            html = subgraph.produce_visualisation()
            html_path = graphs_dir / f"{subgraph.label}.html"
            html_path.write_text(html)
            relative_html_path = Path(f"graphs/{subgraph.label}.html")
            subgraph.path = relative_html_path

    @staticmethod
    def _produce_index_file(
        outdir: Path, graphs: ListOfGraphs[BlackholeGraph], objects_description: str
    ) -> None:
        sorted_graphs = graphs.get_graphs_sorted_by_size()
        index_src = Templates.read_template("index_template")

        visualisation_links = []
        for graph in sorted_graphs:
            visualisation_links.append(
                f'<a href="{graph.path}" target="_blank">View {graph.description} <br/>'
            )

        with open(outdir / "index.html", "w") as index_fh:
            for line in index_src:
                line = line.replace("<communities_links_tag>", "\n".join(visualisation_links))
                line = line.replace("<objects_description>", objects_description)
                print(line, file=index_fh)

        OutputProducer.copy_libs(outdir)

    @staticmethod
    def copy_libs(outdir: Path) -> None:
        shutil.copytree(get_libs_dir(), outdir / "libs", dirs_exist_ok=True)



import json
import math
import pickle
from abc import abstractmethod
from pathlib import Path
from typing import Any, Generator, Optional, TextIO, Type, TypeVar, cast

import networkx as nx

from plasnet.ColorPicker import ColorPicker
from plasnet.Templates import Templates
from plasnet.utils import DistanceTags

BaseGraphType = TypeVar("BaseGraphType", bound="BaseGraph")


class BaseGraph(nx.Graph):  # type: ignore
    """
    Class to represent a base class to concentrate common methods
    between the different types of graphs.
    """

    # def fix_node_to_subcommunity_attributes(self, node_to_subcommunity, blackhole_plasmids):
    #     for node, attrs in self.nodes.items():
    #         if node in node_to_subcommunity:
    #             attrs["color"] = ColorPicker.get_color_given_index(node_to_subcommunity[node])
    #         else:
    #             attrs["color"] = ColorPicker.get_default_color()
    #
    #         if node in blackhole_plasmids:
    #             attrs["shape"] = "star"
    #             attrs["is_blackhole"] = True
    #         else:
    #             attrs["is_blackhole"] = False

    def __init__(self, graph: Optional[nx.Graph] = None, label: str = "") -> None:
        super().__init__(graph)
        self._label = label
        self._path: Optional[Path] = None  # path to html file for visualisation of this graph

    @property
    def label(self) -> str:
        return self._label

    @property
    def path(self) -> Optional[Path]:
        return self._path

    @path.setter
    def path(self, path: Path) -> None:
        self._path = path

    @property
    def description(self) -> str:
        return f"{self.label} ({self.number_of_nodes()} nodes, {self.number_of_edges()} edges)"

    def _get_node_color(self, node: str) -> str:
        return ColorPicker.get_default_color()

    def _get_node_shape(self, node: str) -> str:
        return "circle"

    def _add_special_node_attributes(self, node: str, attrs: dict[str, Any]) -> None:
        ...

    def remove_plasmids(self, plasmids_to_be_removed: list[str]) -> None:
        self.remove_nodes_from(plasmids_to_be_removed)

    def fix_node_attributes(self) -> None:
        for node, attrs in self.nodes.items():
            attrs["color"] = self._get_node_color(node)
            attrs["shape"] = self._get_node_shape(node)
            self._add_special_node_attributes(node, attrs)

    def fix_edge_attributes(self) -> None:
        for edge, attrs in self.edges.items():
            distances = (
                attrs.get(DistanceTags.SplitDistanceTag.value, math.nan),
                attrs.get(DistanceTags.TypeDistanceTag.value, math.nan),
            )
            filtered_distances = filter(lambda x: not math.isnan(x), distances)
            string_distances = map(str, filtered_distances)
            distance_label = " / ".join(string_distances)
            attrs["d_lbl"] = distance_label

    def get_induced_components(self, nodes: list[str]) -> Generator["BaseGraph", None, None]:
        subgraph = self.subgraph(nodes)
        yield from (
            BaseGraph(subgraph.subgraph(component))
            for component in nx.connected_components(subgraph)
        )

    TIME_LIMIT_FOR_SMALL_GRAPHS = 1000
    TIME_LIMIT_FOR_LARGE_GRAPHS = 10000

    def get_simulation_time(self) -> int:
        is_a_small_enough_graph = self.number_of_nodes() <= 5 or self.number_of_edges() <= 10
        if is_a_small_enough_graph:
            return self.TIME_LIMIT_FOR_SMALL_GRAPHS
        else:
            return self.TIME_LIMIT_FOR_LARGE_GRAPHS

    def produce_visualisation(self) -> str:
        self.fix_node_attributes()
        self.fix_edge_attributes()
        visualisation_src = Templates.read_template("visualisation_template")

        graph_as_cy_dict = nx.cytoscape_data(self)
        elements_as_cy_json = json.dumps(graph_as_cy_dict["elements"])

        # [CRITICAL] TODO: improve this horrible performance
        libs_relative_path = self._get_libs_relative_path()
        samples_selectors = self._get_samples_selectors_HTML()
        filters = self._get_filters_HTML()
        custom_buttons = self._get_custom_buttons_HTML()

        final_html_lines = []
        for line in visualisation_src:
            line = line.replace("<libs_relative_path>", libs_relative_path)
            line = line.replace("<samples_selectors>", samples_selectors)
            line = line.replace("<elements_tag>", elements_as_cy_json)
            line = line.replace("<movementThreshold>", str(self.number_of_edges()))
            line = line.replace("<maxSimulationTime>", str(self.get_simulation_time()))
            line = line.replace("<filters_tag>", filters)
            line = line.replace("<custom_buttons_tag>", custom_buttons)
            final_html_lines.append(line)

        return "\n".join(final_html_lines)

    @abstractmethod
    def _get_libs_relative_path(self) -> str:
        ...

    @abstractmethod
    def _get_samples_selectors_HTML(self) -> str:
        ...

    @abstractmethod
    def _get_filters_HTML(self) -> str:
        ...

    @abstractmethod
    def _get_custom_buttons_HTML(self) -> str:
        ...

    def save(self, filepath: Path) -> None:
        with open(filepath, "wb") as fh:
            pickle.dump(self, fh)

    @classmethod
    def load(cls: Type[BaseGraphType], filepath: Path) -> BaseGraphType:
        with open(filepath, "rb") as fh:
            graph = pickle.load(fh)
            return cast(BaseGraphType, graph)

    def write_classification(self, typing_fh: TextIO) -> None:
        for node in self.nodes:
            typing_fh.write(f"{node}\t{self.label}\n")



from plasnet.list_of_graphs import ListOfGraphs
from plasnet.sample_graph import SampleGraph


class SampleGraphs(ListOfGraphs[SampleGraph]):
    def get_graphs_sorted_by_size(self) -> "SampleGraphs":
        nb_of_hit_samples = map(SampleGraph.get_number_of_hit_samples, self)

        graphs_and_nb_of_hit_samples = [
            (graph, nb_of_hit_samples) for graph, nb_of_hit_samples in zip(self, nb_of_hit_samples)
        ]

        graphs_and_nb_of_hit_samples_sorted = sorted(
            graphs_and_nb_of_hit_samples, key=lambda x: x[1], reverse=True
        )
        sorted_graphs = [graph for graph, _ in graphs_and_nb_of_hit_samples_sorted]
        return SampleGraphs(sorted_graphs)



from enum import Enum
from pathlib import Path
from typing import Optional

import click
import pandas as pd


def get_plasnet_source_dir() -> Path:
    return Path(__file__).parent


def get_libs_dir() -> Path:
    return get_plasnet_source_dir() / "ext/libs"


class PathlibPath(click.Path):
    def convert(  # type: ignore
        self, value: str, param: Optional[click.core.Parameter], ctx: Optional[click.core.Context]
    ) -> Path:
        return Path(super().convert(value, param, ctx))


DistanceDict = dict[tuple[str, str], float]


class DistanceTags(Enum):
    SplitDistanceTag = "sd"
    TypeDistanceTag = "td"


def distance_df_to_dict(df: pd.DataFrame) -> DistanceDict:
    df_as_dict = {}
    for _, row in df.iterrows():
        df_as_dict[(row["plasmid_1"], row["plasmid_2"])] = row["distance"]
        df_as_dict[(row["plasmid_2"], row["plasmid_1"])] = row["distance"]
    return df_as_dict



# TODO: refactor

from typing import Optional

import networkx as nx

from plasnet.blackhole_graph import BlackholeGraph
from plasnet.ColorPicker import ColorPicker
from plasnet.subcommunities import Subcommunities
from plasnet.subcommunity_graph import SubcommunityGraph
from plasnet.utils import DistanceDict, DistanceTags


class CommunityGraph(BlackholeGraph):
    def __init__(
        self,
        graph: Optional[nx.Graph] = None,
        blackhole_connectivity_threshold: int = 0,
        edge_density: float = 0.0,
        label: str = "",
    ):
        super().__init__(graph, blackhole_connectivity_threshold, edge_density, label)
        self._node_to_colour: dict[str, str] = {}

    def _get_node_color(self, node: str) -> str:
        return self._node_to_colour.get(node, ColorPicker.get_default_color())

    @staticmethod
    def _get_node_to_subcommunity(subcommunities: list[set[str]]) -> dict[str, int]:
        node_to_subcommunity = {}
        for subcommunity_index, subcommunity in enumerate(subcommunities):
            for node in subcommunity:
                node_to_subcommunity[node] = subcommunity_index
        return node_to_subcommunity

    def _fix_small_subcommunities(
        self, subcommunities: list[set[str]], small_subcommunity_size_threshold: int
    ) -> list[set[str]]:
        # sort subcommunities by size so that we can safely move
        # smaller subcommunities into larger ones
        subcommunities = sorted(subcommunities, key=lambda subcommunity: len(subcommunity))
        node_to_subcommunity = self._get_node_to_subcommunity(subcommunities)

        for subcommunity_idx, subcommunity in enumerate(subcommunities):
            subcommunity_is_too_small = len(subcommunity) <= small_subcommunity_size_threshold
            if subcommunity_is_too_small:
                subcommunity_neighbours = set(
                    neighbor for node in subcommunity for neighbor in self.neighbors(node)
                ) - set(subcommunity)
                subcommunity_has_no_neighbors = len(subcommunity_neighbours) == 0
                if subcommunity_has_no_neighbors:
                    continue

                candidate_subcommunities = set(
                    node_to_subcommunity[neighbor] for neighbor in subcommunity_neighbours
                )
                largest_subcommunity_size = max(
                    len(subcommunities[subcommunity]) for subcommunity in candidate_subcommunities
                )
                small_will_be_integrated_into_large = largest_subcommunity_size >= len(subcommunity)
                if small_will_be_integrated_into_large:
                    largest_subcommunity_idx = next(
                        filter(
                            lambda idx: len(subcommunities[idx]) == largest_subcommunity_size,
                            candidate_subcommunities,
                        )
                    )
                    subcommunities[largest_subcommunity_idx].update(subcommunity)
                    subcommunities[subcommunity_idx] = set()
        return subcommunities

    def split_graph_into_subcommunities(
        self, small_subcommunity_size_threshold: int
    ) -> Subcommunities:
        subcommunities_nodes: list[set[str]] = list(
            nx.community.asyn_lpa_communities(
                G=self, weight=DistanceTags.TypeDistanceTag.value, seed=42
            )
        )
        subcommunities_nodes = self._fix_small_subcommunities(
            subcommunities_nodes, small_subcommunity_size_threshold
        )

        subcommunities = []
        for subcommunity_index, subcommunity_nodes in enumerate(subcommunities_nodes):
            colour = ColorPicker.get_color_given_index(subcommunity_index)

            subcommunity = SubcommunityGraph(
                self.subgraph(subcommunity_nodes),
                self._blackhole_connectivity_threshold,
                self._edge_density,
                label=f"{self.label}_subcommunity_{subcommunity_index}",
                colour=colour,
            )
            subcommunities.append(subcommunity)

            for node in subcommunity_nodes:
                self._node_to_colour[node] = colour

        return Subcommunities(subcommunities)

    def _get_libs_relative_path(self) -> str:
        return ".."

    def _get_samples_selectors_HTML(self) -> str:
        return ""

    def add_typing_distances(self, distance_dict: DistanceDict) -> None:
        for edge in self.edges:
            self.edges[edge][DistanceTags.TypeDistanceTag.value] = distance_dict[edge]

    def filter_by_distance(self, distance_threshold: float) -> None:
        edges_to_remove = []
        for edge in self.edges:
            edge_should_be_removed = (
                self.edges[edge][DistanceTags.TypeDistanceTag.value] > distance_threshold
            )
            if edge_should_be_removed:
                edges_to_remove.append(edge)

        self.remove_edges_from(edges_to_remove)

    def recolour_nodes(self, other_community: "CommunityGraph") -> None:
        for node in self.nodes:
            self._node_to_colour[node] = other_community._node_to_colour[node]



# TODO: refactor
from plasnet.community_graph import CommunityGraph
from plasnet.list_of_graphs import ListOfGraphs
from plasnet.utils import DistanceDict


class Communities(ListOfGraphs[CommunityGraph]):
    def add_typing_distances(self, distance_dict: DistanceDict) -> None:
        for community in self:
            community.add_typing_distances(distance_dict)

    def filter_by_distance(self, distance_threshold: float) -> None:
        for community in self:
            community.filter_by_distance(distance_threshold)

    def recolour_nodes(self, other_communities: "Communities") -> None:
        for my_community, other_community in zip(self, other_communities):
            my_community.recolour_nodes(other_community)



import logging
from copy import deepcopy
from pathlib import Path
from typing import cast

import click
import pandas as pd

from plasnet import __version__
from plasnet.communities import Communities
from plasnet.output_producer import OutputProducer
from plasnet.plasmid_graph import PlasmidGraph
from plasnet.sample_graph import SampleGraph
from plasnet.sample_graphs import SampleGraphs
from plasnet.subcommunities import Subcommunities
from plasnet.utils import PathlibPath, distance_df_to_dict

logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")


@click.group()
@click.version_option(version=__version__)
def cli() -> None:
    pass


@click.command(
    help="Creates and split a plasmid graph into communities",
    epilog="""
\b
Creates and split a plasmid graph into communities.
The plasmid graph is defined by plasmid and distance files.

\b
The plasmid file is a tab-separated file with one column describing all plasmids in the dataset.
Example of such file:
plasmid
AP024796.1
AP024825.1
CP012142.1
CP014494.1
CP019149.1
CP021465.1
CP022675.1
CP024687.1
CP026642.1
CP027485.1

\b
The distances file is a tab-separated file with 3 columns: plasmid_1, plasmid_2, distance.
plasmid_1 and plasmid_2 are plasmid names, and distance is a float between 0 and 1.
The distance threshold is the minimum distance value for two plasmids to be considered connected.
Example of such file:
plasmid_1       plasmid_2       distance
AP024796.1      AP024825.1      0.8
AP024796.1      CP012142.1      0.5
AP024796.1      CP014494.1      0.3
AP024796.1      CP019149.1      0.0
AP024796.1      CP021465.1      0.0
AP024796.1      CP022675.1      1.0
AP024796.1      CP024687.1      0.0
AP024796.1      CP026642.1      0.5
AP024796.1      CP027485.1      0.8
""",  # noqa: E501
)
@click.argument("plasmids", type=PathlibPath(exists=True))
@click.argument("distances", type=PathlibPath(exists=True))
@click.argument("output-dir", type=PathlibPath(exists=False))
@click.option("--distance-threshold", "-d", type=float, default=0.5, help="Distance threshold")
@click.option(
    "--bh-connectivity",
    "-b",
    type=int,
    default=10,
    help="Minimum number of connections a plasmid need to be considered a blackhole plasmid",
)
@click.option(
    "--bh-neighbours-edge-density",
    "-e",
    type=float,
    default=0.2,
    help="Maximum number of edge density between blackhole plasmid neighbours to "
    "label the plasmid as blackhole",
)
@click.option(
    "--output-plasmid-graph",
    "-p",
    is_flag=True,
    help="Also outputs the full, unsplit, plasmid graph",
)
def split(
    plasmids: Path,
    distances: Path,
    output_dir: Path,
    distance_threshold: float,
    bh_connectivity: int,
    bh_neighbours_edge_density: float,
    output_plasmid_graph: bool,
) -> None:
    visualisations_dir = output_dir / "visualisations"
    logging.info(f"Creating plasmid graph from {plasmids} and {distances}")
    plasmid_graph = PlasmidGraph.build(plasmids, distances, distance_threshold)

    if output_plasmid_graph:
        logging.info("Producing full plasmid graph visualisation")
        OutputProducer.produce_graph_visualisation(
            plasmid_graph, visualisations_dir / "single_graph" / "single_graph.html"
        )

    logging.info("Splitting plasmid graph into communities")
    communities = plasmid_graph.split_graph_into_communities(
        bh_connectivity, bh_neighbours_edge_density
    )

    logging.info("Producing communities visualisation")
    OutputProducer.produce_communities_visualisation(
        communities, visualisations_dir / "communities"
    )

    logging.info("Serialising objects")
    objects_dir = output_dir / "objects"
    objects_dir.mkdir(parents=True, exist_ok=True)
    plasmid_graph.save(objects_dir / "plasmid_graph.pkl")
    communities.save(objects_dir / "communities.pkl")
    communities.save_graph_as_text(objects_dir / "communities.txt")
    communities.save_classification(objects_dir / "communities.tsv", "plasmid\tcommunity")

    logging.info("All done!")


@click.command(
    help="Type the communities of a previously split plasmid graph into subcommunities or types",
    epilog="""
\b
Type the communities of a previously split plasmid graph into subcommunities or types.
This typing is based on running an asynchronous label propagation algorithm on the previously identified communities.
This algorithm is implemented in the networkx library, and relies on a given distance file.
This distance file should be a more precise and careful distance function than the one used to split the graph into communities.
For example, you could use gene jaccard distance to split the graph and the DCJ-indel distance to type the communities.
See https://github.com/iqbal-lab-org/pling for a tool to compute gene jaccard and DCJ-indel distances.

\b
The first file, describing the communities, is a pickle file (.pkl) that can be found in <split_out_dir>/objects/communities.pkl,
where <split_out_dir> is the output dir of the split command.

\b
The distances file is a tab-separated file with 3 columns: plasmid_1, plasmid_2, distance.
plasmid_1 and plasmid_2 are plasmid names, and distance is a float number.
The distance threshold is the minimum distance value for two plasmids to be considered connected.
Example of such file:
plasmid_1       plasmid_2       distance
AP024796.1      AP024825.1      4
AP024796.1      CP012142.1      10
AP024796.1      CP014494.1      20
AP024796.1      CP019149.1      1
AP024796.1      CP021465.1      0
AP024796.1      CP022675.1      50
AP024796.1      CP024687.1      1000
AP024796.1      CP026642.1      20
AP024796.1      CP027485.1      1
""",  # noqa: E501
)
@click.argument("communities-pickle", type=PathlibPath(exists=True))
@click.argument("distances", type=PathlibPath(exists=True))
@click.argument("output-dir", type=PathlibPath(exists=False))
@click.option("--distance-threshold", "-d", type=float, default=4, help="Distance threshold")
@click.option(
    "--small-subcommunity-size-threshold",
    type=int,
    default=4,
    help="Subcommunities with size up to this parameter will be joined to "
    "neighbouring larger subcommunities",
)
def type(
    communities_pickle: Path,
    distances: Path,
    output_dir: Path,
    distance_threshold: float,
    small_subcommunity_size_threshold: int,
) -> None:
    logging.info(f"Loading communities from {communities_pickle}")
    communities = cast(Communities, Communities.load(communities_pickle))

    output_dir.mkdir(parents=True, exist_ok=True)

    logging.info(f"Loading typing distances from {distances}")
    distance_df = pd.read_csv(distances, sep="\t")
    distance_dict = distance_df_to_dict(distance_df)

    logging.info("Adding typing distance in plasmid graph")
    communities.add_typing_distances(distance_dict)

    logging.info("Backing up communities before applying distance filter")
    original_communities = deepcopy(communities)

    logging.info("Applying distance filter")
    communities.filter_by_distance(distance_threshold)

    logging.info("Typing communities (i.e. splitting them into subcommunities)")
    all_subcommunities = Subcommunities()
    for community in communities:
        community.remove_blackhole_plasmids()
        subcommunities = community.split_graph_into_subcommunities(
            small_subcommunity_size_threshold
        )
        all_subcommunities.extend(subcommunities)

    logging.info("Producing communities visualisations")
    original_communities.recolour_nodes(communities)
    OutputProducer.produce_communities_visualisation(
        original_communities, output_dir / "visualisations/communities"
    )

    logging.info("Producing subcommunities visualisations")
    OutputProducer.produce_subcommunities_visualisation(
        all_subcommunities, output_dir / "visualisations/subcommunities"
    )

    logging.info("Serialising objects")
    objects_dir = output_dir / "objects"
    objects_dir.mkdir(parents=True, exist_ok=True)
    original_communities.save(objects_dir / "communities.pkl")
    all_subcommunities.save(objects_dir / "subcommunities.pkl")
    all_subcommunities.save_classification(objects_dir / "typing.tsv", "plasmid\ttype")

    logging.info("All done!")


@click.command(
    help="Add sample hits annotations on top of previously identified subcommunities or types",
    epilog="""
\b
Add sample hits annotations on top of previously identified subcommunities or types.


\b
The first file, describing the subcommunities, is a pickle file (.pkl) that can be found in <type_out_dir>/objects/subcommunities.pkl,
where <type_out_dir> is the output dir of the type command.

\b
The sample-hits file is a tab-separated file with 2 columns: sample, plasmid.
These columns are self-explanatory and identifies the plasmids present in each sample.
Example of such file:
sample              plasmid
cpe001_trim_ill     NZ_CP006799.1
cpe001_trim_ill     NZ_CP028929.1
cpe002_trim_ill     NZ_CP079159.1
cpe005_trim_ill     NZ_CP006799.1
cpe005_trim_ill     NZ_CP079676.1
cpe010_trim_ill     NZ_CP028929.1
cpe020_trim_ill     NZ_CP006799.1
cpe020_trim_ill     NZ_CP079676.1
cpe021_trim_ill     NZ_CP006799.1
""",  # noqa: E501
)
@click.argument("subcommunities-pickle", type=PathlibPath(exists=True))
@click.argument("sample-hits", type=PathlibPath(exists=True))
@click.argument("output-dir", type=PathlibPath(exists=False))
def add_sample_hits(
    subcommunities_pickle: Path,
    sample_hits: Path,
    output_dir: Path,
) -> None:
    logging.info(f"Loading subcommunities from {subcommunities_pickle}")
    subcommunities = cast(Subcommunities, Subcommunities.load(subcommunities_pickle))

    logging.info(f"Loading sample hits from {sample_hits}")
    sample_hits_df = pd.read_csv(sample_hits, sep="\t")

    output_dir.mkdir(parents=True, exist_ok=True)

    logging.info("Producing sample graphs")
    sample_graphs = SampleGraphs()
    for subcommunity in subcommunities:
        sample_plasmid_for_this_subcommunity = sample_hits_df[
            sample_hits_df["plasmid"].isin(subcommunity.nodes)
        ]
        sample_graph = SampleGraph.from_subcommunity_graph(
            subcommunity, sample_plasmid_for_this_subcommunity
        )
        sample_graphs.append(sample_graph)

    logging.info("Producing sample graphs visualisations")
    OutputProducer.produce_subcommunities_visualisation(
        sample_graphs, output_dir / "visualisations/sample_graphs"
    )

    logging.info("All done!")


# Add commands to the main group
cli.add_command(split)
cli.add_command(type)
cli.add_command(add_sample_hits)


def main() -> None:
    """Entry point for the application script"""
    cli()



import sys

"""
Version has unique source in pyproject.toml.
importlib fetches version from distribution metadata files
(in dist-info or egg-info dirs).
From Python 3.8, importlib_metadata is in standard library as importlib.metadata.
"""
if sys.version_info >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata

__version__: str = metadata.version("plasnet")

from .plasnet_main import main

__all__ = ["main"]



import pickle
from pathlib import Path
from typing import Generator, cast

from plasnet.base_graph import BaseGraphType


class ListOfGraphs(list[BaseGraphType]):
    def save(self, filepath: Path) -> None:
        with open(filepath, "wb") as fh:
            pickle.dump(self, fh)

    @classmethod
    def load(cls, filepath: Path) -> "ListOfGraphs[BaseGraphType]":
        with open(filepath, "rb") as fh:
            graphs = pickle.load(fh)
            return cast(ListOfGraphs[BaseGraphType], graphs)

    def _get_each_graph_as_list_of_nodes_in_text_format(
        self,
    ) -> Generator[str, None, None]:
        for graph in self:
            yield " ".join(graph)

    def save_graph_as_text(self, filepath: Path) -> None:
        with open(filepath, "w") as fh:
            for graph_as_text in self._get_each_graph_as_list_of_nodes_in_text_format():
                print(graph_as_text, file=fh)

    def save_classification(self, filepath: Path, header: str) -> None:
        with open(filepath, "w") as fh:
            print(header, file=fh)
            for subgraph in self:
                subgraph.write_classification(fh)

    def get_graphs_sorted_by_size(self) -> "ListOfGraphs[BaseGraphType]":
        return ListOfGraphs(sorted(self, key=lambda graph: graph.number_of_nodes(), reverse=True))



import random

RNG = random.Random(42)


class ColorPicker:
    colours = [
        "darkred",
        "indianred",
        "fuchsia",
        "hotpink",
        "darkgreen",
        "chartreuse",
        "darkblue",
        "blue",
        "aqua",
        "darkviolet",
        "mediumpurple",
        "dimgrey",
        "darkorange",
        "maroon",
        "yellow",
        "orange",
        "teal",
    ]
    basic_colours = [
        "red",
        "magenta",
        "green",
        "blue",
        "black",
        "orange",
        "maroon",
        "cyan",
    ]
    RNG.shuffle(colours)

    @staticmethod
    def get_color_given_index(i: int) -> str:
        return ColorPicker.colours[i % len(ColorPicker.colours)]

    @staticmethod
    def get_default_color() -> str:
        return "black"



from typing import Optional

import networkx as nx

from plasnet.blackhole_graph import BlackholeGraph
from plasnet.ColorPicker import ColorPicker


class SubcommunityGraph(BlackholeGraph):
    def __init__(
        self,
        graph: Optional[nx.Graph] = None,
        blackhole_connectivity_threshold: int = 0,
        edge_density: float = 0.0,
        label: str = "",
        colour: str = ColorPicker.get_default_color(),
    ):
        super().__init__(graph, blackhole_connectivity_threshold, edge_density, label)
        self._colour = colour

    def _get_libs_relative_path(self) -> str:
        return ".."

    def _get_samples_selectors_HTML(self) -> str:
        return ""

    def _get_node_color(self, node: str) -> str:
        return self._colour



import logging
from typing import Any, Optional

import networkx as nx

from plasnet.base_graph import BaseGraph


class BlackholeGraph(BaseGraph):
    """
    This is a class that recognises and labels blackhole plasmids.
    """

    def __init__(
        self,
        graph: Optional[nx.Graph] = None,
        blackhole_connectivity_threshold: int = 0,
        edge_density: float = 0.0,
        label: str = "",
    ):
        super().__init__(graph, label)
        self._blackhole_connectivity_threshold = blackhole_connectivity_threshold
        self._edge_density = edge_density
        self._blackhole_plasmids = self._get_blackhole_plasmids()

    def _get_node_shape(self, node: str) -> str:
        if node in self._blackhole_plasmids:
            return "star"
        return "circle"

    def _add_special_node_attributes(self, node: str, attrs: dict[str, Any]) -> None:
        if node in self._blackhole_plasmids:
            attrs["is_blackhole"] = True
        else:
            attrs["is_blackhole"] = False

    def _get_blackhole_plasmids(self) -> list[str]:
        blackhole_plasmids_in_graph = []
        for node in self.nodes:
            if self.degree(node) >= self._blackhole_connectivity_threshold:
                neighbors = list(self.neighbors(node))
                subgraph = nx.induced_subgraph(self, neighbors)
                nb_of_edges_between_neighbours = subgraph.number_of_edges()
                max_nb_of_edges_between_neighbours = (len(neighbors) * (len(neighbors) - 1)) // 2
                edge_rate = nb_of_edges_between_neighbours / max_nb_of_edges_between_neighbours
                if edge_rate <= self._edge_density:
                    blackhole_plasmids_in_graph.append(node)
                    logging.debug(f"{node} is a blackhole plasmid, REMOVED")
                else:
                    logging.debug(
                        f"{node} is highly connected but does not connect "
                        f"unrelated plasmids, not removed"
                    )
        return blackhole_plasmids_in_graph

    def get_nb_of_blackhole_plasmids(self) -> int:
        return len(self._blackhole_plasmids)

    def remove_blackhole_plasmids(self) -> None:
        self.remove_nodes_from(self._blackhole_plasmids)

    def _get_filters_HTML(self) -> str:
        nb_of_black_holes = len(self._get_blackhole_plasmids())
        return (
            f'<label for="hide_blackholes">'
            f"Hide blackhole plasmids ({nb_of_black_holes} present)"
            f"</label>"
            f'<input type="checkbox" id="hide_blackholes" name="hide_blackholes"><br/>'
        )

    def _get_custom_buttons_HTML(self) -> str:
        return '<div><input type="submit" value="Redraw" onclick="redraw()"></div>'

    @property
    def description(self) -> str:
        description = super().description
        blackholes_detected = len(self._blackhole_plasmids) > 0
        if blackholes_detected:
            description += " - WARNING: BLACKHOLE SPOTTED!"
        return description



from pathlib import Path

from plasnet.utils import get_plasnet_source_dir


class Templates:
    @staticmethod
    def get_templates_dir() -> Path:
        return get_plasnet_source_dir() / "ext/templates"

    @staticmethod
    def read_template(template_name: str) -> list[str]:
        template_filepath = Templates.get_templates_dir() / (template_name + ".html")
        with open(template_filepath) as template_fh:
            template_src = template_fh.readlines()
        return list(map(lambda line: line.strip("\n"), template_src))
